# main
# Author:       Everett Harding, Luke Buquicchio
# Date Created: 1/19/18
# Description:
#    main file for running experiments. This file handles setting any and all parameters specified by the user
# and applying defaults. It manages running and saving the overall experiments, including:
#   - initial random splitting of data
#   - calling model build functions with proper parameters
#   - training models (both subsetNets and adapter)
#   - saving models & Results
#
import csv
import threading
import random
import datetime
import os

import numpy as np

from Utilities import data_utils
from Utilities import ensemble_utils
from Utilities import multithreaded_ensemble
from Utilities import io_utils
from adapter import Adapter
from sklearn.model_selection import StratifiedKFold
from keras import utils as k_utils

<<<<<<< HEAD:model_tester.py
from balance_data import upsample_minority

=======
>>>>>>> 6c4326b18f62b351f6a3a3f2d53bab589c9ca03d:Legacy/model_tester.py
# Parameters/flags
# - actiavtion functions for ensemble networks
# - initializtion of nodes in ensemble networks
# - # of nets in ensemble
# - # of layers in ensemble nets
# - dimensions of layers in ensemble nets
# - maximum threads for ensemble training
#
# - activation functions for adapter network
# - intialization of nodes in adapter network
# - # of layers in adapter network
# - dimension of layers in adapter network
#
# - percent of features in subset
# - percent training/testing split
#
# - number of epochs
# - cross-validation folds

def main():
    """
        ensemble training flow:
            - set parameters (# layers, shape of layers, min feature set size, training subset size)
            - split training subsets into folds, cross validate
            - record:
                - success by feature (x_5 was predicted with an average error of .75 over 3 nets)
                - average success (error of .62 over all nets)
                - discrete features get confusion matricies and f1 scores
                    - how to integrate with averages?
                    - can hold separate, compare discrete classifications to continuous predictions
        :param params: The parameters of the main method. Can be only filename as of now.

    """

    np.set_printoptions(linewidth=1000)
    parser = io_utils.get_arg_parser()
    params = parser.parse_args()

    # find and load the data
    filename = params.dataset
    data = np.loadtxt(filename, delimiter=',', skiprows=1)

    labels = data[:, data.shape[1]-1]
    data = np.delete(data, data.shape[1]-1, 1)

    if filename == 'Data/UCI_Credit_Card.csv':
        data = np.delete(data, 0, 1)

    data = data_utils.normalize(data)
    x_train, x_test, y_train, y_test = data_utils.split_data(data, labels, params.train_test_split)

    # handle other parameters
    # ensemble_act = params.ensemble_activation
    # ensemble_init = params.ensemble_initialization
    ensemble_size = params.ensemble_size

    ensemble_layers = params.ensemble_layers
    # ensemble_epochs = params.ensemble_epochs

    # adapter_act = params.adapter_activation
    # adapter_init = params.adapter_initialization

    adapter_layers = params.adapter_layers
    # adapter_epochs = params.adapter_epochs

    subset_split = params.subset_split

    print("EL:", ensemble_layers)
    print("AL:", adapter_layers)

    # Set the seed
    seed = 42
    np.random.seed(seed)
    random.seed(seed)

    # rows = data.shape[0]

    # use the final column in the dataset as y
    layers_dict = ensemble_utils.build_hidden_layer_dict(ensemble_layers)
    ensemble = ensemble_utils.generate_ff_ensemble(train_data=x_train,
                                                   seed=seed,
                                                   num_nets=ensemble_size,
                                                   layers_dict=layers_dict,
                                                   column_train_test_split=subset_split)

    ensemble_results = []
    for network in ensemble:
        result = {}
        result['model_results'] = network.model.evaluate(x_test[:, network.feature_mask], y_test, verbose=0)
        result['label'] = network.target_feature
        result['mask'] = network.feature_mask
        ensemble_results.append(result)
    # retrieve the penultimate layer outputs from the data
    new_data = ensemble_utils.model_output_matrix(ensemble, x_train)

    print("New Data Shape:", new_data.shape)
    # mask = np.random.rand(rows) < train_test_split

    adapter_layers_dict = ensemble_utils.build_hidden_layer_dict(adapter_layers)

    # k-fold cross validation
    kfold = StratifiedKFold(n_splits=params.k_folds, shuffle=True, random_state=seed)
    cv_scores = []

    # split the data, fit model, and evaluate.
    print("Cross-validating adapter")
    for train, test in kfold.split(new_data, y_train):
        print(".", end="")
        # adapt the output
        cols = new_data.shape[1]
        adapter = Adapter(cols, hidden_units=adapter_layers_dict)
        adapter.fit(new_data[train], labels[train], epochs=150)
        scores = adapter.evaluate(new_data[test], labels[test])
        cv_scores.append(scores)

    with open(filename, 'r') as file:
        all_fieldnames = np.array(csv.DictReader(file).fieldnames)
        if filename == 'Data/UCI_Credit_Card.csv':
            fieldnames = all_fieldnames[1:-1]

        print("saving scores")
        io_utils.simulation_output_file(params, cv_scores, ensemble_results, fieldnames)
        print("scores saved")

    return cv_scores


<<<<<<< HEAD:model_tester.py
def optimizable_main(args):
    """
    This main will only optimize the Adapter and work with a much larger, but much
    more naive ensemble.
    List of arguments to run the program with:
                    idx 0: ensemble
                    idx 1: adapter_activation
                    idx 2: adapter_layer1: The size of the first layer in the adapter.
                    idx 3: adapter_layer2: The size of the second layer in the adapter.
                    idx 4: adapter_layer3: The size of the third layer in the adapter.
                    idx 5: adapter_loss: The loss function for the Adapter.
    :return: The testing accuracy of the Adapter.
    """
    data = args[0]


    if not os.path.exists('./Models'):
        os.makedirs('./Models')

    os.chdir('./Models')

    now = datetime.datetime.now()
    now_formatted = str(now.strftime("%Y-%m-%d-%H:%M:%S"))

    if not os.path.exists(now_formatted):
        os.makedirs(now_formatted)
    os.chdir(now_formatted)

    ensemble = args[1]
    adapter_activation = args[2]
    adapter_layer1 = int(args[3])
    adapter_layer2 = int(args[4])
    adapter_layer3 = int(args[5])
    adapter_loss = args[6]

    adapter_layers = [adapter_layer1]
    if not adapter_layer2 == 0:
        adapter_layers.append(adapter_layer2)
    if not adapter_layer3 == 0:
        adapter_layers.append(adapter_layer3)

    # Please note that the following three lines are unique to the UCI data and
    # should be changed according to the column meanings of the data set.
    # ################################################
    #labels = data[:, data.shape[1] - 1]
    #data = np.delete(data, data.shape[1] - 1, 1)
    #data = np.delete(data, 0, 1)

    #data = data_utils.normalize(data)

    #data, labels = data_utils.bag_data(data, labels)

    X, y = upsample_minority(data)
    x_train, x_test, y_train, y_test = data_utils.split_data(X, y, .70)
    # ################################################

    zeros = 0
    ones = 0
    for label in y_test:
        if label == 0:
            zeros += 1
        else:
            ones += 1

    ratio = ones/zeros

    # np.random.seed(seed)
    # random.seed(seed)

    # retrieve the penultimate layer outputs from the data
    new_train_data = ensemble_utils.model_output_matrix(ensemble, x_train)

    adapter_layers_dict = ensemble_utils.build_hidden_layer_dict(adapter_layers)
    cols = new_train_data.shape[1]
    out_shape = 1

    if adapter_loss == 'categorical_crossentropy':
        out_shape = 2
        y_train = k_utils.to_categorical(y_train.astype(int))
        y_test = k_utils.to_categorical(y_test.astype(int))
    elif adapter_loss == 'categorical_hinge':
        out_shape = 1
    elif adapter_loss == 'binary_crossentropy':
        out_shape = 1

    avg_ensemble_loss = 0.0
    avg_ensemble_accuracy = 0.0
    for model in ensemble:
        input_data = x_test[:, model.feature_mask]
        input_label = x_test[:, model.target_feature]
        avg_ensemble_accuracy += model.evaluate(input_data, input_label)[1]
        avg_ensemble_loss += model.evaluate(input_data, input_label)[0]
    avg_ensemble_accuracy = avg_ensemble_accuracy / len(ensemble)
    avg_ensemble_loss = avg_ensemble_loss / len(ensemble)

    adapter = Adapter(cols,
                      hidden_units=adapter_layers_dict,
                      activation=adapter_activation,
                      loss=adapter_loss,
                      num_gpus=0,
                      verbose=0,
                      out_shape=out_shape)

    adapter.fit(new_train_data, y_train, epochs=128)

    new_test_data = ensemble_utils.model_output_matrix(ensemble, x_test)
    score = adapter.evaluate(new_test_data, y_test)

    u = 0
    for model in ensemble:
        model.save("ensemble_" + str(u) + ".h5")
        u += 1

    adapter.save("adapter" + ".h5")

    del ensemble
    del adapter

    os.chdir('../..')

    return score, ratio, avg_ensemble_accuracy, avg_ensemble_loss


=======
>>>>>>> 6c4326b18f62b351f6a3a3f2d53bab589c9ca03d:Legacy/model_tester.py
def optimizable_main_jl(args):
    """
    Juan Luis's idea for aggregating penultimate layers from the ensemble in steps
    to compare their testing accuracies.
    :param args: List of arguments to run the program with:
                    idx 0: num_nets
                    idx 1: adapter_activation
                    idx 2: adapter_initializer
                    idx 3: adapter_layer1: The size of the first layer in the adapter.
                    idx 4: adapter_layer2: The size of the second layer in the adapter.
                    idx 5: adapter_layer3: The size of the third layer in the adapter.
    :return: The testing accuracy of the Adapter.
    """
    data = np.loadtxt('../Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)

    if not os.path.exists('../Models'):
        os.makedirs('../Models')

    os.chdir('../Models')

    now = datetime.datetime.now()
    now_formatted = str(now.strftime("%Y-%m-%d-%H:%M:%S"))

    if not os.path.exists(now_formatted):
        os.makedirs(now_formatted)
    os.chdir(now_formatted)

    num_nets = args[0]
    adapter_activation = int(args[1])
    # adapter_initializer = int(args[2])
    adapter_layer1 = int(args[3])
    adapter_layer2 = int(args[4])
    adapter_layer3 = int(args[5])

    adapter_layers = [adapter_layer1]
    if not adapter_layer2 == 0:
        adapter_layers.append(adapter_layer2)
    if not adapter_layer3 == 0:
        adapter_layers.append(adapter_layer3)

    adapter_activations = ['relu', 'sigmoid']
    # adapter_initializers = ['uniform', 'normal', 'logistic']

    # Please note that the following three lines are unique to the UCI data and
    # should be changed according to the column meanings of the data set.
    labels = data[:, data.shape[1] - 1]
    data = np.delete(data, data.shape[1] - 1, 1)
    data = np.delete(data, 0, 1)

    data = data_utils.normalize(data)
    x_train, x_test, y_train, y_test = data_utils.split_data(data, labels, .70)
    ensemble_size = num_nets

    subset_split = .7
    seed = 42
    np.random.seed(seed)
    random.seed(seed)

    # use the final column in the dataset as y
    layers_dict = ensemble_utils.build_hidden_layer_dict([10])
    ensemble = ensemble_utils.generate_ff_ensemble(train_data=x_train,
                                                   seed=seed,
                                                   num_nets=ensemble_size,
                                                   layers_dict=layers_dict,
                                                   column_train_test_split=subset_split,
                                                   num_gpus=0)

    random.shuffle(ensemble)

    print("Compiling & Fitting Model  Adapter")

    # retrieve the penultimate layer outputs from the data
    new_train_data = ensemble_utils.model_output_matrix(ensemble, x_train)
    new_test_data = ensemble_utils.model_output_matrix(ensemble, x_test)

    adapter_layers_dict = ensemble_utils.build_hidden_layer_dict(adapter_layers)

    for n in range(int(num_nets/2)):

        print("Using the first " + str(num_nets/2 + n) + " ensemble members")

        new_train_data_copy = new_train_data[:, 0:int(((num_nets/2) * 10) + (n * 10))]
        new_test_data_copy = new_test_data[:, 0:int(((num_nets/2) * 10) + (n * 10))]

        print("Train shape ", new_train_data_copy.shape)
        print("Test shape  ", new_test_data_copy.shape)

        cols = new_train_data_copy.shape[1]

        adapter = Adapter(cols,
                          hidden_units=adapter_layers_dict,
                          activation=adapter_activations[adapter_activation],
                          num_gpus=0)
        adapter.fit(new_train_data_copy, y_train, epochs=100)

        score = adapter.evaluate(new_test_data_copy, y_test)
        print(score)

    print("Finished adapter training & eval: ", 1.0 - float(score[0]))

    u = 0
    for model in ensemble:
        model.save("ensemble_" + str(u) + ".h5")
        u += 1

    adapter.save("adapter" + ".h5")

    del ensemble
    del adapter

    os.chdir('..')

    return 1.0 - float(score[1])


def threaded_optimizable_main(args):
    """
    This is the multithreaded main to be used by a an optimization routine.
    Please note that this function should no longer be used.
    :param args:
    :return:
    """
    num_nets = args[0]
    adapter_activation = int(args[1])
    adapter_initializer = int(args[2])
    adapter_layer1 = int(args[3])
    adapter_layer2 = int(args[4])
    adapter_layer3 = int(args[5])

    adapter_layers = [adapter_layer1]
    if not adapter_layer2 == 0:
        adapter_layers.append(adapter_layer2)
    if not adapter_layer3 == 0:
        adapter_layers.append(adapter_layer3)

    adapter_activations = ['relu', 'softmax']
    adapter_initializers = ['random_uniform', 'normal', 'logistic']

    data = np.loadtxt('../Data/UCI_Credit_Card.csv', delimiter=',', skiprows=1)

    # Please note that the following three lines are unique to the UCI data and
    # should be changed according to the column meanings of the data set.
    labels = data[:, data.shape[1] - 1]
    data = np.delete(data, data.shape[1] - 1, 1)
    data = np.delete(data, 0, 1)

    data = data_utils.normalize(data)
    x_train, x_test, y_train, y_test = data_utils.split_data(data, labels, .70)

    subset_split = .7
    seed = 42
    np.random.seed(seed)
    random.seed(seed)

    # use the final column in the dataset as y
    layers_dict = ensemble_utils.build_hidden_layer_dict([10])

    for worker in range(int(num_nets)):
        multithreaded_ensemble.thread_queue.put(worker)

    for z in range(int(num_nets/4)):
        t = threading.Thread(target=multithreaded_ensemble.ensemble_threader(train_data=x_train,
                                                                             column_train_test_split=subset_split,
                                                                             num_gpus=0,
                                                                             layers_dict=layers_dict,))
        t.daemon = True
        t.start()

    multithreaded_ensemble.thread_queue.join()

    # retrieve the penultimate layer outputs from the data
    new_data_train = ensemble_utils.model_output_matrix(multithreaded_ensemble.ensemble, x_train)
    new_data_test = ensemble_utils.model_output_matrix(multithreaded_ensemble.ensemble, x_test)

    adapter_layers_dict = ensemble_utils.build_hidden_layer_dict(adapter_layers)

    cols = new_data_train.shape[1]
    adapter = Adapter(cols,
                      hidden_units=adapter_layers_dict,
                      activation=adapter_activations[adapter_activation],
                      initializer=adapter_initializers[adapter_initializer],
                      num_gpus=0)
    adapter.fit(new_data_train, y_train, epochs=100)
    score = adapter.evaluate(new_data_test, y_test)

    return 1.0 - float(score[0])


if __name__ == "__main__":
    threaded_optimizable_main([10, 1, 0, 100, 50, 10])
