#!/bin/env python3
# network_build_functions
# Author:       Everett Harding
# Date Created: 2/20/18
# Description: 
# TODO:
import numpy as np
np.set_printoptions(linewidth=1000)
import random
import sys
sys.path.extend(['/work/ejharding/MQP-ZHAW'])
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # or any {'0', '1', '2'}
import psutil
from keras.backend import clear_session
from keras.utils import np_utils
from Utilities.data_utils import load_uci_set, normalize, downsample_data
from keras.layers import Input, Dense, Dropout
from keras.models import Model
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import LabelBinarizer


def check_categorical(feature_idx):
    categorical_mask = [(False, -1),(True,2),(True,6), (True,3), (False, -1), (True,10),
                        (True,10), (True,10),(True,10), (True,10),(True,10),(False, -1),
                        (False, -1),(False, -1),(False, -1),(False, -1),(False, -1),(False, -1),
                        (False, -1),(False, -1),(False, -1),(False, -1),(False, -1)]
    return categorical_mask[feature_idx]


def build_classification_ffnnetwork(in_shape, num_categories, max_hidden_layers=3, max_units=46, activation='relu', initalizer='uniform', dropout_prob=-1., opt='sgd', loss='categorical_crossentropy'):
    # track all the parameters for this network
    params = {
        'input_shape': in_shape,
        'activation': activation,
        'initalizer': initalizer,
        'optimizer': opt,
        'loss': loss,
        'num_categories': num_categories
    }

    num_layers = np.random.randint(1, max_hidden_layers + 1)

    # build the input layer
    inputs = Input((in_shape,))
    first_layer_units = np.random.randint(2, max_units)
    m = Dense(first_layer_units)(inputs)
    params['input_layer_units'] = first_layer_units

    # build the guaranteed hidden layer
    first_hidden_units = np.random.randint(2, max_units)
    m = Dense(first_hidden_units, activation=activation, kernel_initializer=initalizer)(m)
    params['hidden1_units'] = first_hidden_units

    # add optional dropout
    if np.random.rand() < dropout_prob:
        rate1 = np.random.rand()
        m = Dropout(rate1)(m)
        params['dropout1_rate'] = rate1

    # add optional addtional hidden layers and dropouts
    if num_layers > 1:
        second_layer_units = np.random.randint(2, max_units)
        m = Dense(second_layer_units, activation=activation, kernel_initializer=initalizer)(m)
        params['hidden2_units'] = second_layer_units

    if np.random.rand() < dropout_prob:
        rate2 = np.random.rand()
        m = Dropout(rate2)(m)
        params['dropout2_rate'] = rate2

    if num_layers > 2:
        third_layer_units = np.random.randint(2, max_units)
        m = Dense(third_layer_units, activation=activation, kernel_initializer=initalizer)(m)
        params['hidden3_units'] = third_layer_units

    m = Dense(num_categories, activation='softmax')(m)

    # now put the whole thing together
    model = Model(inputs, m)
    model.compile(optimizer=opt, loss=loss, metrics=['accuracy'])

    return model, params


def build_feed_forward_network(in_shape, max_hidden_layers=3, max_units=46, activation='relu', initalizer='uniform', dropout_prob=-1., opt='sgd', loss='mse'):
    # track all the parameters for this network
    params = {
        'input_shape': in_shape,
        'activation': activation,
        'initalizer': initalizer,
        'optimizer': opt,
        'loss': loss
    }

    num_layers = np.random.randint(1, max_hidden_layers+1)

    # build the input layer
    inputs = Input((in_shape,))
    first_layer_units = np.random.randint(2, max_units)
    m = Dense(first_layer_units)(inputs)
    params['input_layer_units'] = first_layer_units

    # build the guaranteed hidden layer
    first_hidden_units = np.random.randint(2, max_units)
    m = Dense(first_hidden_units, activation=activation, kernel_initializer=initalizer)(m)
    params['hidden1_units'] = first_hidden_units

    # add optional dropout
    if np.random.rand() < dropout_prob:
        rate1 = np.random.rand()
        m = Dropout(rate1)(m)
        params['dropout1_rate'] = rate1

    # add optional addtional hidden layers and dropouts
    if num_layers > 1:
        second_layer_units = np.random.randint(2, max_units)
        m = Dense(second_layer_units,  activation=activation, kernel_initializer=initalizer)(m)
        params['hidden2_units'] = second_layer_units

    if np.random.rand() < dropout_prob:
        rate2 = np.random.rand()
        m = Dropout(rate2)(m)
        params['dropout2_rate'] = rate2

    if num_layers > 2:
        third_layer_units = np.random.randint(2, max_units)
        m = Dense(third_layer_units,  activation=activation, kernel_initializer=initalizer)(m)
        params['hidden3_units'] = third_layer_units

    m = Dense(1, activation='linear')(m)

    # now put the whole thing together
    model = Model(inputs, m)
    model.compile(optimizer=opt, loss=loss, metrics=['accuracy']) # accuracy included here to make sure we play nice with the categorical ones

    return model, params


def build_ensemble(num_nets, train,  net_params, keep_feature_threshhold=.1):
    """
    given a set of maximum parameters, returns a list of compiled and trained ensemble networks
    :return: the entire ensemble of networks, compiled and trained
    """
    ensemble = {}
    ensemble_size = 0
    num_cols = train.shape[1]
    unused_labels = list(range(0, num_cols))

    # set up all the parameters we'll need for building the networks
    batch_size = net_params['batch_size']

    while ensemble_size < num_nets:
        # If all of the columns have been used, reuse them all.
        if not unused_labels:
            unused_labels = list(range(0, num_cols))

        # Randomly pick one of those to be the new label
        label_index = unused_labels[random.randint(0, len(unused_labels) - 1)]
        # Remove the label from the index
        unused_labels.remove(label_index)
        # Retrieve all of the data from the column index
        net_training_label_unencoded = train[:, label_index]

        # determine if the selected feature is categorical and if so, how many categories?
        is_cat, num_categories = check_categorical(label_index)

        if is_cat:
            # set up
            lb = LabelBinarizer()
            lb.fit(net_training_label_unencoded)
            net_training_label = lb.transform(net_training_label_unencoded)

            if lb.y_type_ == 'binary':
                net_training_label = np_utils.to_categorical(net_training_label)

            num_categories = len(lb.classes_)
        else:
            net_training_label = net_training_label_unencoded

        # Generate a list of booleans indicating columns to keep (True) and to ignore (False)
        mask = np.random.rand(num_cols - 1) < keep_feature_threshhold
        while not mask.any():
            mask = np.random.rand(num_cols - 1) < keep_feature_threshhold


        # Boolean array that includes a FALSE at the original index of the temp label
        total_mask = np.insert(mask, label_index, False)
        # now use that to get a view of the columns we want to use to train
        net_training_columns = train[:, total_mask]

        # the categorical nets get built differently from the regressors, so plan accordingly...
        if is_cat:
            # making the classifiers means you need to save the binarizer as well
            net, params = build_classification_ffnnetwork(net_training_columns.shape[1], num_categories)
            params['lb'] = lb
        else:
            # making the regressors is less taxing, just make them
            net, params = build_feed_forward_network(net_training_columns.shape[1])

        # time to train the net
        net.fit(net_training_columns, net_training_label, batch_size=batch_size, epochs=net_params['epochs'], verbose=0)


        if label_index in ensemble.keys():
            ensemble[label_index].append((net, params, total_mask, label_index))
        else:
            ensemble[label_index] = [(net, params, total_mask, label_index)]

        ensemble_size += 1

    return ensemble


def encode(data, ensemble):
    """

    :param data:
    :param ensemble:
    :return:
    """
    encoded_matrix = None

    for entry in ensemble.values():
        for net, params, mask, label in entry:
            ins = net.input
            outs = net.get_layer(index=-2).output
            inter_model = Model(ins, outs)

            transformed = inter_model.predict(data[:, mask])
            if np.isnan(transformed).any():
                continue

            if encoded_matrix is None:
                encoded_matrix = transformed
            else:
                encoded_matrix = np.append(encoded_matrix, transformed, axis=1)

    return encoded_matrix


def build_adapter(in_shape, layers, activation='relu', initalizer='uniform', dropout_prob=-1., opt='adam'):

    params = {
        'in_shape': in_shape
    }

    ins = Input((in_shape,))

    m = Dense(layers[0], activation=activation, kernel_initializer=initalizer)(ins)
    m = Dropout(dropout_prob)(m)
    layer_num = 0
    for units in layers[1:]:
        layer_num+=1

        m = Dense(units, activation=activation, kernel_initializer=initalizer)(m)
        m = Dropout(dropout_prob)(m)
        key = 'hidden' + str(layer_num)
        params[key] = units

    m = Dense(2, activation='softmax')(m)

    model = Model(ins, m)
    model.compile(optimizer=opt, loss='binary_crossentropy', metrics=['accuracy'])

    return model, params


def train_framework(train_features, train_labels, params):

    ens_size = params['ens_size']
    ens_batch_size = params['ens_batch_size']
    ens_epochs = params['ens_epochs']

    adpt_batch_size = params['adpt_batch_size']
    adpt_epochs = params['adpt_epochs']
    layers = params['adpt_layers']

    train_f = train_features
    train_l = train_labels

    ens = build_ensemble(ens_size, train_f, {'batch_size': ens_batch_size, 'epochs': ens_epochs})
    transformed = encode(train_features, ens)

    adapter_in = transformed.shape[1]
    adapter, adapter_params = build_adapter(adapter_in, layers, dropout_prob=.5)

    adapter.fit(transformed, train_l, batch_size=adpt_batch_size, epochs=adpt_epochs, verbose=0)

    adapter_params['batch_size'] = adpt_batch_size
    adapter_params['epochs'] = adpt_epochs

    return adapter, ens, adapter_params


def evaluate_framework(test_f, test_l, adapter, ensemble):

    # print("Evaluating Framework")
    evaluation_string = ""
    ensemble_results = {}
    for entry in ensemble.values():
        avg_loss = 0
        for net, params, mask, label in entry:
            # unpack the tuple and use it to get some info
            is_cat, num_cat = check_categorical(label)
            if is_cat:
                lb = params['lb']
                # lb.fit(test_f[:,label])
                labels = lb.transform(test_f[:,label])

                if lb.y_type_ == 'binary':
                    labels = np_utils.to_categorical(labels)

            else:
                labels = test_f[:,label]

            result = net.evaluate(test_f[:, mask], labels, verbose=0)

            if is_cat:
                evaluation_string += "Target %4d:\tAccy: %.4f\tStd Dev: %.4f\n" % (label, result[1], test_f[:, label].std())
            else:
                evaluation_string += "Target %4d:\tLoss: %.4f\tStd Dev: %.4f\n" % (label, result[0], test_f[:, label].std())

            # print("Target %4d:\tLoss: %.4f\tStd Dev: %.4f" % (label, result, test_f[:, label].std()), flush=True)
            avg_loss += result[0]

            if label in ensemble_results.keys():
                ensemble_results[label].append((result, params, mask))
            else:
                ensemble_results[label] = [(result, params, mask)]
        evaluation_string += "Avg Loss: %.4f\n\n" % (avg_loss / len(entry))

    test_f = encode(test_f, ensemble)
    evaluation_string += "Aggregated Hidden Dim Size:" + str(test_f.shape[1]) + '\n'
    result = adapter.evaluate(test_f, test_l, verbose=0)
    evaluation_string += "Loss: %.3f\t Accuracy: %.3f\n\n================================\n" % (result[0], result[1])
    print(evaluation_string)
    return result, ensemble_results


def normalize(data):
    cont_mask = [True,False,False,False,True,False,False,False,False,False,False,True,True,True,True,True,True,True,True,True,True,True,True]
    max = np.max(data[:,cont_mask])
    min = np.min(data[:,cont_mask])
    data[:,cont_mask] = (data[:, cont_mask] - min) / (max - min)

    return data

if __name__ == '__main__':

    raw, labels = load_uci_set()
    features = normalize(raw)
    x_vault, y_vault = features[int(features.shape[0] * .9):], labels[int(labels.shape[0] * .9):]
    features, labels = features[:int(features.shape[0] * .9)], labels[:int(labels.shape[0] * .9)]
    features, labels = downsample_data(features, labels)

    framework = {
        'ens_size': np.random.randint(10,50),
        'ens_batch_size': np.random.randint(5,10),
        'ens_epochs': np.random.randint(1,10),
        'adpt_batch_size': np.random.randint(1000,1050),
        'adpt_epochs': np.random.randint(10,15),
        'adpt_layers': [np.random.randint(1,x * 20) for x in range(np.random.randint(1,6),0,-1)]
    }
    np.random.seed(7)
    skf = StratifiedKFold(n_splits=10, shuffle=True)

    y = np_utils.to_categorical(labels)
    res = 0

    for train_index, test_index in skf.split(features, labels):
        x_train, x_test = features[train_index], features[test_index]
        y_train, y_test = y[train_index], y[test_index]

        # get trained versions of the entire framework
        adapter, ensemble, adapter_params = train_framework(x_train, y_train, framework)
        # now get results from testing over the whole framework
        adapter_result, ens_params = evaluate_framework(x_test, y_test, adapter, ensemble)

        # print info on memory usage for Turing Debugging
        process = psutil.Process(os.getpid())
        before = process.memory_info()[1]
        print("mem before:", before, flush=True)
        clear_session()
        process = psutil.Process(os.getpid())
        after = process.memory_info()[1]
        print("mem after:", after, flush=True)

        res += adapter_result[1]

        res_string = 'Overall Accuracy: %.3f\tOverall Loss: %.3f\n' % (adapter_result[1], adapter_result[0])
        for key in adapter_params.keys():
            res_string += str(key) + ': %d\n' % adapter_params[key]

        for key in ens_params:
            res_string += str(key) + ':\n'
            for result, params, mask in ens_params[key]:
                res_string += '\tMask: ' + str(mask) + '\n'
                res_string += '\tParameters: ' + str(params) + '\n'
        print("History", adapter.history.history, flush=True)
        print(res_string, flush=True)

    print("Cross-Validation Accuracy: %.2f" %(res/10.), flush=True)



