# subsetPredictionNet
# Author: Everett Harding & Luke Buquicchio

# Date Created: 1/24/18
# Description:
#   This file will serve as the utilities file for all of the Neural Networks in the ensemble as well as the adapter.

from sklearn.utils import resample
from sklearn.metrics import confusion_matrix, precision_score, recall_score
from matplotlib import pyplot as plt
from sklearn.metrics import f1_score, cohen_kappa_score
import pandas as pd
import numpy as np
import datetime
import os


def load_uci_set():
    """
    :return: the UCI Credit Card Default dataset as a numpy array
    """

    cwd = os.getcwd()
    if cwd[-8:] == "MQP-ZHAW":
        filename = "Data/UCI_Credit_Card.csv"
    else:
        filename = "../Data/UCI_Credit_Card.csv"

    data = np.loadtxt(filename, delimiter=',', skiprows=1)

    x = data[:, 1:-1]
    y = data[:, -1]
    return x, y


def normalize(data, reg_columns=None):
    """
    A simple function to normalize the data. (Value / max value in the column)
    :param data: The data to be normalized.
    :param reg_columns: The mask of columns of the data set that are regression.
    :return: The normalized data.
    """
    max = np.max(data[:, reg_columns])
    min = np.min(data[:, reg_columns])
    data[:, reg_columns] = (data[:, reg_columns] - min) / (max - min)

    return data


def downsample_data(data, labels):
    """
    given the entire dataset and labels returns a set with equal parts 0 and 1 labels
    :param data:
    :param labels:
    :return: views of data and labels containing the bagged samples from each
    """
    sorted_idx = labels.argsort()

    cutoff = 0
    while labels[sorted_idx[cutoff]] == 0.:
        cutoff += 1

    zeros = np.random.choice(sorted_idx[:cutoff], len(sorted_idx[cutoff:]))
    ones = sorted_idx[cutoff:]

    bagged = np.concatenate((zeros, ones))

    return data[bagged], labels[bagged]


def split_data(data, labels, split):
    """
    A simple function to split the data into training data and labels and testing data and labels.
    The reason that data and labels exist as separate parameters it to be easily adapt to new data sets.
    :param data: The data to be split.
    :param labels: The labels to be split.
    :param split: The decimal representation of the percent split between training and testing.
    :return: the 4 different ndarrays.
    """
    rows = data.shape[0]  # The number of rows in the data set

    # mask for splitting selected columns and labels into training and testing
    full_set_mask = np.random.rand(rows) < split
    x_train = data[full_set_mask]
    x_test = data[~full_set_mask]
    y_train = labels[full_set_mask]
    y_test = labels[~full_set_mask]

    return x_train, x_test, y_train, y_test


def generate_df(data):
    predictors = [
        'ID', 'LIMIT_BAL', 'SEX', 'EDUCATION', 'MARRIAGE', 'AGE', 'PAY_0',
        'PAY_2', 'PAY_3', 'PAY_4', 'PAY_5', 'PAY_6', 'BILL_AMT1', 'BILL_AMT2',
        'BILL_AMT3', 'BILL_AMT4', 'BILL_AMT5', 'BILL_AMT6', 'PAY_AMT1',
        'PAY_AMT2', 'PAY_AMT3', 'PAY_AMT4', 'PAY_AMT5', 'PAY_AMT6', 'default'
    ]
    return pd.DataFrame(data, columns=predictors)


def upsample_minority(data):
    df = generate_df(data)

    # Separate majority and minority classes
    df_majority = df[df.default == 0]
    df_minority = df[df.default == 1]

    # Upsample minority class
    df_minority_upsampled = resample(
        df_minority,
        replace=True,  # sample with replacement
        n_samples=23364,  # to match majority class
        random_state=123)  # reproducible results

    df_upsampled = pd.concat([df_majority, df_minority_upsampled])

    # Separate input features (X) and target variable (y)
    y = df_upsampled.default
    X = df_upsampled.drop(['default', 'ID'], axis=1)

    y = y.as_matrix(columns=None)
    X = X.as_matrix(columns=None)

    return X, y


def downsample_majority(data):
    df = generate_df(data)

    # Separate majority and minority classes
    df_majority = df[df.default == 0]
    df_minority = df[df.default == 1]

    # Downsample majority class
    df_majority_downsampled = resample(
        df_majority,
        replace=False,  # sample without replacement
        n_samples=6636,  # to match minority class
        random_state=123)  # reproducible results

    # Combine minority class with downsampled majority class
    df_downsampled = pd.concat([df_majority_downsampled, df_minority])

    y = df_downsampled.default
    X = df_downsampled.drop(['default', 'ID'], axis=1)

    y = y.as_matrix(columns=None)
    X = X.as_matrix(columns=None)

    return X, y


def synthetic_minority_sampling(data):
    return None


def additional_metrics(model, y_test, x_test):
    y_pred = model.predict(x_test)

    print('confusion matrix:')
    conf_matrix = confusion_matrix(y_test.argmax(axis=1), y_pred.argmax(axis=1))
    print(conf_matrix)
    print('precision score:')
    prec_score = precision_score(y_test.argmax(axis=1), y_pred.argmax(axis=1))
    print(prec_score)
    print('recall score:')
    rec_score = recall_score(y_test.argmax(axis=1), y_pred.argmax(axis=1))
    print(rec_score)
    print('f1 score:')
    f_score = f1_score(y_test.argmax(axis=1), y_pred.argmax(axis=1))
    print(f_score)
    print('cohen kappa score:')
    kappa = cohen_kappa_score(y_test.argmax(axis=1), y_pred.argmax(axis=1))
    print(kappa)
    #print('Area Under ROC:')
    #print(roc_auc_score(y_test.argmax(axis=1), y_pred.argmax(axis=1)))

    return (conf_matrix, prec_score, rec_score, f_score, kappa)


def save_metrics_file(model, y_test, x_test):
    y_pred = model.predict(x_test)

    if not os.path.exists('benchmark_results'):
        os.makedirs('benchmark_results')

    os.chdir('benchmark_results')
    now = datetime.datetime.now()
    filename = str(now.strftime("%Y-%m-%d-%H:%M:%S"))

    with open(filename, 'a') as results_file:
        print('confusion matrix:', file=results_file)
        conf_matrix = confusion_matrix(y_test.round(), y_pred.round())
        print(conf_matrix, file=results_file)
        prec_score = precision_score(y_test.round(), y_pred.round())
        print('precision score: %f' % prec_score, file=results_file)
        rec_score = recall_score(y_test.round(), y_pred.round())
        print('recall score: %f' % rec_score, file=results_file)
        f_score = f1_score(y_test.round(), y_pred.round())
        print('f1 score: %f' % f_score, file=results_file)
        kappa = cohen_kappa_score(y_test.round(), y_pred.round())
        print('cohen kappa score: %f' % kappa, file=results_file)


def collect_data_benchmark_files():
    precision_score_data = []
    precision_str = 'precision score: '
    recall_score_data = []
    recall_str = 'recall score: '
    f1_score_data = []
    fscore_str = 'f1 score: '
    kappa_data = []
    kappa_str = 'cohen kappa score: '

    for filename in os.listdir(os.getcwd()):
        with open(filename, 'r') as current_file:
            line = current_file.readline()
            while line:
                if precision_str in line:
                    precision_score_data.append(line.split(precision_str, 1)[1])
                elif recall_str in line:
                    recall_score_data.append(line.split(recall_str, 1)[1])
                elif fscore_str in line:
                    f1_score_data.append(line.split(fscore_str, 1)[1])
                elif kappa_str in line:
                    kappa_data.append(line.split(kappa_str, 1)[1])

    precision_score_data = list(map(float, precision_score_data))
    recall_score_data = list(map(float, recall_score_data))
    f1_score_data = list(map(float, f1_score_data))
    kappa_data = list(map(float, kappa_data))

    return precision_score_data, recall_score_data, f1_score_data, kappa_data


def plot_precision_ensemble_size(precision_score, ensemble_size):
    plt.plot(ensemble_size, precision_score)
    plt.xlabel('Ensemble Size')
    plt.ylabel('Precision Score')
    plt.title('Ensemble Size vs Precision Score')
    plt.legend()
    plt.show()

def plot_recall_ensemble_size(recall_score, ensemble_size):
    plt.plot(ensemble_size, recall_score)
    plt.xlabel('Ensemble Size')
    plt.ylabel('Recall Score')
    plt.title('Ensemble Size vs Recall Score')
    plt.legend()
    plt.show()

def plot_fscore_ensemble_size(fscore, ensemble_size):
    plt.plot(ensemble_size, fscore)
    plt.xlabel('Ensemble Size')
    plt.ylabel('F1 Score')
    plt.title('Ensemble Size vs F1 Score')
    plt.legend()
    plt.show()

def plot_kappa_ensemble_size(kappa_score, ensemble_size):
    plt.plot(ensemble_size, kappa_score)
    plt.xlabel('Ensemble Size')
    plt.ylabel('Kappa Score')
    plt.title('Ensemble Size vs Kappa Score')
    plt.legend()
    plt.show()
