# subsetPredictionNet
# Author: Everett Harding, Luke Buquicchio, & Juan Luis Herrero

# Date Created: 1/24/18
# Description:
#   This file will serve as the utilities file for all of the Neural Networks in the ensemble as well as the adapter.

import random
import keras as ks
import numpy as np

from keras import utils as k_utils
from keras.layers import Input, Dense, Dropout
from keras.models import Model
from sklearn.preprocessing import LabelBinarizer

import Utilities.data_utils as du
import ensembleNetworks as en


def build_classification_ff_network(in_shape, num_categories, max_hidden_layers=3, max_units=46, activation='relu', initalizer='uniform', dropout_prob=-1., opt='sgd', loss='categorical_crossentropy'):
    # track all the parameters for this network
    params = {
        'input_shape': in_shape,
        'activation': activation,
        'initalizer': initalizer,
        'optimizer': opt,
        'loss': loss,
        'num_categories': num_categories
    }

    num_layers = np.random.randint(1, max_hidden_layers + 1)

    # build the input layer
    inputs = Input((in_shape,))
    first_layer_units = np.random.randint(2, max_units)
    m = Dense(first_layer_units)(inputs)
    params['input_layer_units'] = first_layer_units

    # build the guaranteed hidden layer
    first_hidden_units = np.random.randint(2, max_units)
    m = Dense(first_hidden_units, activation=activation, kernel_initializer=initalizer)(m)
    params['hidden1_units'] = first_hidden_units

    # add optional dropout
    if np.random.rand() < dropout_prob:
        rate1 = np.random.rand()
        m = Dropout(rate1)(m)
        params['dropout1_rate'] = rate1

    # add optional addtional hidden layers and dropouts
    if num_layers > 1:
        second_layer_units = np.random.randint(2, max_units)
        m = Dense(second_layer_units, activation=activation, kernel_initializer=initalizer)(m)
        params['hidden2_units'] = second_layer_units

    if np.random.rand() < dropout_prob:
        rate2 = np.random.rand()
        m = Dropout(rate2)(m)
        params['dropout2_rate'] = rate2

    if num_layers > 2:
        third_layer_units = np.random.randint(2, max_units)
        m = Dense(third_layer_units, activation=activation, kernel_initializer=initalizer)(m)
        params['hidden3_units'] = third_layer_units

    m = Dense(num_categories, activation='softmax')(m)

    # now put the whole thing together
    model = Model(inputs, m)
    model.compile(optimizer=opt, loss=loss, metrics=['accuracy'])

    return model, params


def build_feed_forward_network(in_shape, max_hidden_layers=3, max_units=46, activation='relu', initalizer='uniform', dropout_prob=-1., opt='sgd', loss='mse'):
    # track all the parameters for this network
    params = {
        'input_shape': in_shape,
        'activation': activation,
        'initalizer': initalizer,
        'optimizer': opt,
        'loss': loss
    }

    num_layers = np.random.randint(1, max_hidden_layers+1)

    # build the input layer
    inputs = Input((in_shape,))
    first_layer_units = np.random.randint(2, max_units)
    m = Dense(first_layer_units)(inputs)
    params['input_layer_units'] = first_layer_units

    # build the guaranteed hidden layer
    first_hidden_units = np.random.randint(2,max_units)
    m = Dense(first_hidden_units, activation=activation, kernel_initializer=initalizer)(m)
    params['hidden1_units'] = first_hidden_units

    # add optional dropout
    if np.random.rand() < dropout_prob:
        rate1 = np.random.rand()
        m = Dropout(rate1)(m)
        params['dropout1_rate'] = rate1

    # add optional addtional hidden layers and dropouts
    if num_layers > 1:
        second_layer_units = np.random.randint(2, max_units)
        m = Dense(second_layer_units,  activation=activation, kernel_initializer=initalizer)(m)
        params['hidden2_units'] = second_layer_units

    if np.random.rand() < dropout_prob:
        rate2 = np.random.rand()
        m = Dropout(rate2)(m)
        params['dropout2_rate'] = rate2

    if num_layers > 2:
        third_layer_units = np.random.randint(2,max_units)
        m = Dense(third_layer_units,  activation=activation, kernel_initializer=initalizer)(m)
        params['hidden3_units'] = third_layer_units

    m = Dense(1, activation='linear')(m)

    # now put the whole thing together
    model = Model(inputs, m)
    model.compile(optimizer=opt, loss=loss, metrics=['accuracy']) # accuracy included here to make sure we play nice with the categorical ones


    return model, params


def build_ensemble(num_nets, train, net_params, cat_columns, keep_feature_threshhold=.7):
    """
    given a set of maximum parameters, returns a list of compiled and trained ensemble networks
    :return: the entire ensemble of networks, compiled and trained
    """
    ensemble = {}
    ensemble_size = 0
    num_cols = train.shape[1]
    unused_labels = list(range(0, num_cols))

    # set up all the parameters we'll need for building the networks
    batch_size = net_params['batch_size']

    while ensemble_size < num_nets:
        # If all of the columns have been used, reuse them all.
        if not unused_labels:
            unused_labels = list(range(0, num_cols))

        # Randomly pick one of those to be the new label
        label_index = unused_labels[random.randint(0, len(unused_labels) - 1)]
        # Remove the label from the index
        unused_labels.remove(label_index)
        # Retrieve all of the data from the column index
        net_training_label_unencoded = train[:, label_index]

        # determine if the selected feature is categorical and if so, how many categories?
        is_cat = cat_columns[label_index]

        if is_cat:
            # set up
            lb = LabelBinarizer()
            lb.fit(net_training_label_unencoded)
            net_training_label = lb.transform(net_training_label_unencoded)

            if lb.y_type_ == 'binary':
                net_training_label = k_utils.to_categorical(net_training_label)

            num_categories = len(lb.classes_)
        else:
            net_training_label = net_training_label_unencoded

        # Generate a list of booleans indicating columns to keep (True) and to ignore (False)
        mask = np.random.rand(num_cols - 1) < keep_feature_threshhold
        while not mask.any():
            mask = np.random.rand(num_cols - 1) < keep_feature_threshhold


        # Boolean array that includes a FALSE at the original index of the temp label
        total_mask = np.insert(mask, label_index, False)

        # The columns used for training the neural nets
        remaining_cols = train
        remaining_cols = np.delete(remaining_cols, label_index, 1)

        # net_training_columns = remaining_cols[:, mask]
        net_training_columns = train[:, total_mask]

        # the categorical nets get built differently from the regressors, so plan accordingly...
        if is_cat:
            # making the classifiers means you need to save the binarizer as well
            net, params = build_classification_ff_network(net_training_columns.shape[1], num_categories)
            params['lb'] = lb
        else:
            # making the regressors is less taxing, just make them
            net, params = build_feed_forward_network(net_training_columns.shape[1])

        # Here's the exciting part: if we bag the data here
        # (AND ONLY HERE) we can be sure that we are not
        # a)  dishonestly datasnooping
        # b)  training on unbalanced data

        # net_training_columns, net_training_label = du.bag_data(net_training_columns, net_training_label_unencoded)
        # time to train the net
        net.fit(net_training_columns, net_training_label, batch_size=batch_size, epochs=net_params['epochs'], verbose=0)


        if label_index in ensemble.keys():
            ensemble[label_index].append((net, params, total_mask, label_index))
        else:
            ensemble[label_index] = [(net, params, total_mask, label_index)]

        ensemble_size += 1

    return ensemble


def encode(data, ensemble):

    encoded_matrix = None

    for entry in ensemble.values():
        for net, params, mask, label in entry:
            ins = net.input
            outs = net.get_layer(index=-2).output
            inter_model = Model(ins, outs)

            transformed = inter_model.predict(data[:, mask])
            if np.isnan(transformed).any():
                continue

            if encoded_matrix is None:
                encoded_matrix = transformed
            else:
                encoded_matrix = np.append(encoded_matrix, transformed, axis=1)

    return encoded_matrix




























































def generate_ff_ensemble(train_data,
                         seed=None,
                         num_nets=25,
                         column_train_test_split=.7,
                         batch_size=100,
                         num_epochs=10,
                         num_gpus=0,
                         layers_dict=None,
                         reg_columns=None,
                         cat_ranges=None):
    """
    The generate_ensemble function takes in all the necessary parameters to build a list of FeedForwardEnsemble nets
    and returns a list of trained neural nets.
    :param train_data: The data set.
    :param seed: The seed for the random number generators. (This allows repeatable results)
    :param num_nets: The number of neural networks in the ensemble.
    :param column_train_test_split: The split between how many columns are used for training and how many are left for
                                    potential labels.
    :param batch_size: The number of rows to process at a time. Note this is system memory dependant.
    :param num_epochs: The number of epochs to be used in the training.
    :param num_gpus: The number of GPUs to distribute the training over.
    :param layers_dict: Dictionary of hidden layers to use in the ensemble networks
                        mapped as: "layer name" : number of units in the layer
    :return ensemble: The list of neural networks that make up the ensemble.
    """
    # Set the seed and shuffle the data
    np.random.seed(seed)
    random.seed(seed)
    np.random.shuffle(train_data)

    cols = train_data.shape[1]  # The number of columns in the data set

    ensemble = []  # The list of nets

    unused_labels = list(range(0, cols-1))  # A list of column indices that hasn't been used

    while len(ensemble) < num_nets:
        # If all of the columns have been used, reuse them all.
        if not unused_labels:
            unused_labels = list(range(0, cols))

        # Randomly pick one of those to be the new label
        label_index = unused_labels[random.randint(0, len(unused_labels) - 1)]
        # Remove the label from the index
        unused_labels.remove(label_index)
        # Retrieve all of the data from the column index
        net_training_label = train_data[:, label_index]

        # Generate a list of booleans indicating columns to keep (True) and to ignore (False)
        mask = np.random.rand(cols-1) < column_train_test_split

        # The columns used for training the neural nets
        remaining_cols = train_data
        remaining_cols = np.delete(remaining_cols, label_index, 1)

        net_training_columns = remaining_cols[:, mask]

        # Boolean array that includes a FALSE at the original index of the temp label
        total_mask = np.insert(mask, label_index, False)

        if reg_columns[label_index]:
            layers_dict = {'inputLayer': net_training_columns.shape[1], 'hidden1': 10, 'hidden2': 4, 'output': 1}
            net = en.FeedForwardEnsembleNet(net_training_columns.shape[1],
                                            layers_dict,
                                            total_mask,
                                            label_index,
                                            name="ENet_" + str(len(ensemble)),
                                            batch_size=batch_size)

        else:
            out_shape = cat_ranges[label_index]
            layers_dict = {'inputLayer': net_training_columns.shape[1], 'hidden1': 10, 'hidden2': 4,
                           'output': out_shape}

            net_training_label = k_utils.to_categorical(net_training_label.astype(int), cat_ranges[label_index])
            net = en.FeedForwardEnsembleNet(net_training_columns.shape[1],
                                            layers_dict,
                                            total_mask,
                                            label_index,
                                            loss="categorical_crossentropy",
                                            name="ENet_" + str(len(ensemble)),
                                            batch_size=batch_size,
                                            initializer='random_uniform')


        if num_gpus:
            # The beginning of the multi-gpu implementation
            parallel_net = ks.utils.multi_gpu_model(net, num_gpus)
            parallel_net.fit(net_training_columns, net_training_label, 10)

            # Add the net to the ensemble
            ensemble.append(parallel_net)

        else:
            net.fit(net_training_columns, net_training_label, num_epoch=num_epochs)
            ensemble.append(net)

    return ensemble


def pen_output(ffn_object, data):
    """
    Creates a new model with the output layer being the penultimate layer.
    Runs predict on that new model.
    Return the output from the penultimate layer.
    :param ffn_object: The feed-forward neural network object.
    :param data: The data set used by the FFNN.
    :return: The penultimate layer output.
    """
    intermediate_layer_model = ks.Model(
        inputs=ffn_object.model.input,
        outputs=ffn_object.model.get_layer(index=-2).output)
    intermediate_output = intermediate_layer_model.predict(data)

    return intermediate_output


def model_output_matrix(models, data):
    """
    Returns a new data set that is the same number of rows as the original
    but the columns are made up of the penultimate layer outputs from the
    ensemble concatinated. I.e. N X (num_nets * penultimate_output_shape)
    :param models: All of the models in the ensemble.
    :param data: The data set used by the FFNN.
    :return: The new data set.
    """
    layer_output_matrix = None

    for model in models:
        input_data = data[:, model.feature_mask]
        layer_output = pen_output(model, input_data)
        ok = True

        for x in layer_output:
            for y in x:
                if np.isnan(y):
                    ok = False

        if layer_output_matrix is None and ok:
            layer_output_matrix = layer_output

        elif ok:
            layer_output_matrix = np.append(layer_output_matrix, layer_output, axis=1)

    return layer_output_matrix


def build_hidden_layer_dict(layer_list, out_shape=1):
    """
    A function to build the dictionary entries for each layer of an ensemble member.
    :param layer_list: A python list of layers in the net.
    :param out_shape: The shape of the output layer.
    :return: The dictionary object.
    """
    layer_dict = {}
    num_hidden = 0

    for layer_units in layer_list:
        num_hidden += 1
        layer_name = "hidden_" + str(num_hidden)
        layer_dict[layer_name] = layer_units

    layer_dict['output'] = out_shape
    return layer_dict
