# subsetPredictionNet
# Author: Everett Harding

# Date Created: 1/24/18
# Description:
#   This file will serve as the utilities file for all of the Neural Networks in the ensemble as well as the adapter.

# import threading
# import time
import numpy as np
import argparse
import datetime
import os

# default values

dataset = 'Data/UCI_Credit_Card.csv'
kfold = 10
tt_split = .8
sub_split = .7
adapter_epochs = 50
adapter_layers = [20, 10]
adapter_init = 'uniform'
adapter_activation = 'relu'
adapter_gpus = 2
max_threads = 5
ensemble_activation = 'relu'
ensemble_epochs = 5
ensemble_layers = [20, 10]
ensemble_init = 'uniform'
ensemble_size = 5


def get_arg_parser():
    # Parameters/flags
    # - activation functions for ensemble networks
    # - initializtion of nodes in ensemble networks
    # - # of nets in ensemble
    # - # of layers in ensemble nets
    # - dimensions of layers in ensemble nets
    # - maximum threads for ensemble training
    #
    # - activation functions for adapter network
    # - intialization of nodes in adapter network
    # - # of layers in adapter network
    # - dimension of layers in adapter network
    #
    # - percent of features in subset
    # - percent training/testing split
    #
    # - number of epochs
    # - cross-validation folds

    parser = argparse.ArgumentParser()

    # add all the feed-forward ensemble layer arguments
    ensemble = parser.add_argument_group('ff_ensemble')
    ensemble.add_argument('-ea', '--ensemble_activation', help='Ensemble Activation Function: specify the activation function to be used in the ensemble networks', default=ensemble_activation)
    ensemble.add_argument('-ei', '--ensemble_initialization', help='Ensemble Initialization Function: specify the initialization function in the layers of the ensemble networks', default=ensemble_init)
    ensemble.add_argument('-es', '--ensemble_size', type=int, help='Ensemble Size: specify the number of neural nets in the ensemble', default=ensemble_size)
    ensemble.add_argument('-el', '--ensemble_layers', type=int, nargs='+', help='EnsembleLayers: number of units in hidden layer(s) of the ensemble networks. One layer is created for each integer value provided', default=ensemble_layers)
    ensemble.add_argument('-mt', '--max_threads', type=int, help='Maximum threads: maximum number of allowed threads for training ensemble', default=max_threads)
    ensemble.add_argument('-ee', '--ensemble_epochs', type=int, help='Number of Epochs: number of epochs allowed when training ensemble networks', default=ensemble_epochs)

    # add all the adapter arguments
    adapter = parser.add_argument_group('adapter')
    adapter.add_argument('-aa', '--adapter_activation', help='Adapter Activation Function: specify the actiavation function to use in the adapter network layers', default=adapter_activation)
    adapter.add_argument('-ai', '--adapter_initialization', help='Adapter Initialization Function: specify the initialization function in the layers of the adapter network', default=adapter_init)
    adapter.add_argument('-al', '--adapter_layers', type=int, nargs='+', help='Adapter Layers: specify the number of units in the hidden layer(s) of the adapter network. One layer is created for each integer value provided', default=adapter_layers)
    adapter.add_argument('-ae', '--adapter_epochs', type=int, help='Number of Epochs: number of epochs allowed when training adapter network', default=adapter_epochs)
    adapter.add_argument('-ag', '--adapter_gpus', type=int, help='Number of GPUs: number of GPUs to distribute the model over', default=adapter_gpus)

    # misc arguments
    misc = parser.add_argument_group('misc')
    misc.add_argument('-s', '--subset_split', type=float, help='Subset Split: Percent of features to be included in the feature subset fed to the ensemble for training', default=sub_split)
    misc.add_argument('-t', '--train_test_split', type=float, help="Train/Test Split: float percent of dataset to be used in the training set. 1 - training percent will form the evaluate percent", default=tt_split)
    misc.add_argument('-k', '--k_folds', type=int, help='K-Folds: number of folds used in cross-validation', default=kfold)

    # the dataset to use
    dataset_args = parser.add_argument_group('data')
    dataset_args.add_argument('-d', '--dataset', help="Dataset: path to a csv file containing the dataset on which to fit", default=dataset)

    return parser


def simulation_output_file(params, cv_scores, ens_results, fieldnames):
    """
    Produces a file with parameters used during the simulation, and
    also outputs the the cross-validated scores from the evaluate.
    :param params: Neural network parameters like input shape and activation function.
    :param cv_scores: Accuracy Results from kfold cross validation.
    :param ens_results: Results from each net in the ensemble.
    :param fieldnames: The names of each field.
    :return: None
    """


    if not os.path.exists('Results'):
        os.makedirs('Results')

    os.chdir('Results')
    now = datetime.datetime.now()
    now_formatted = str(now.strftime("%Y-%m-%d-%H:%M:%S"))
    print('simulation parameters', file=open(now_formatted, 'a'))
    print('==========================', file=open(now_formatted, 'a'))
    print('----ensemble parameters----', file=open(now_formatted, 'a'))
    print(('ensemble activation function: %s' % params.ensemble_activation), file=open(now_formatted,'a'))
    print(('ensemble init function: %s' % params.ensemble_initialization), file=open(now_formatted, 'a'))
    print(('ensemble size: %5d' % params.ensemble_size), file=open(now_formatted, 'a'))
    print(('ensemble layers: %s' % params.ensemble_layers), file=open(now_formatted, 'a'))
    print(('ensemble max threads: %5d' % params.max_threads), file=open(now_formatted, 'a'))
    print(('ensemble epochs: %5d' % params.ensemble_epochs), file=open(now_formatted, 'a'))

    print('----adapter parameters----', file=open(now_formatted, 'a'))
    print(('adapter activation function: %s' % params.adapter_activation), file=open(now_formatted, 'a'))
    print(('adapter init function: %s' % params.adapter_initialization), file=open(now_formatted, 'a'))
    print(('adapter layers: %s' % params.adapter_layers), file=open(now_formatted, 'a'))
    print(('adapter epochs: %5d' % params.adapter_epochs), file=open(now_formatted, 'a'))

    print('----misc parameters----', file=open(now_formatted, 'a'))
    print(('subset split: %.2f%%' % params.subset_split), file=open(now_formatted, 'a'))
    print(('fit evaluate split: %.2f%%' % params.train_test_split), file=open(now_formatted, 'a'))
    print(('K folds: %5d' % params.k_folds), file=open(now_formatted, 'a'))
    print(('dataset: %s' % params.dataset), file=open(now_formatted, 'a'))
    print('==========================', file=open(now_formatted, 'a'))

    count = 0
    for res in ens_results:
        acc = res['model_results'][1]
        target = fieldnames[res['label']]
        field_mask = res['mask']
        names = np.array(fieldnames)[field_mask]
        print('Net: %d \n\tacc: %.2f%% \n\tTarget: %s \n\tColumns: %s' % (count, acc * 100, target, names), file=open(now_formatted, 'a'))
        count += 1
    print('\nsimulation cross-validation Results', file=open(now_formatted, 'a'))
    for score in cv_scores:
        print('acc: %.2f%%' % (score[1] * 100), file=open(now_formatted, 'a'))

    return None
