# multithreaded_ensemble
# Author: Luke Buquicchio

# Date Created: 1/31/18
# Description:
#   This file will serve as the utilities file for all of the Neural Networks in the ensemble as well as the adapter.

import random
import threading
from queue import Queue
import keras as ks
import numpy as np
import ensembleNetworks as en

ensemble = []
thread_queue = Queue()


def ensemble_threader(train_data,
                      column_train_test_split=.7,
                      batch_size=100,
                      num_epochs=10,
                      num_gpus=0,
                      layers_dict=None):
    """
    A multithreaded implementation of the ensemble generator.
    :param train_data: The data used for training.
    :param column_train_test_split: The split between how many columns are used for training and how many are left for
                                    potential labels.
    :param batch_size: The number of rows to process at a time. Note this is system memory dependant.
    :param num_epochs: The number of epochs to be used in the training.
    :param num_gpus: The number of GPUs to distribute the training over.
    :param layers_dict: Dictionary of hidden layers to use in the ensemble networks
                        mapped as: "layer name" : number of units in the layer.
    """

    thread_lock = threading.Lock()
    np.random.shuffle(train_data)
    i = 0

    while True:
        worker = thread_queue.get()
        net = generator_helper(worker,
                               train_data=train_data,
                               id_num=i,
                               column_train_test_split=column_train_test_split,
                               batch_size=batch_size,
                               num_epochs=num_epochs,
                               num_gpus=num_gpus,
                               layers_dict=layers_dict)

        with thread_lock:
            ensemble.append(net)
            i += 1

        thread_queue.task_done()


def generator_helper(worker,
                     train_data,
                     id_num,
                     column_train_test_split=.7,
                     batch_size=100,
                     num_epochs=10,
                     num_gpus=0,
                     layers_dict=None):
    """
    The helper function for the multithreaded implementation of the ensemble generator.
    :param worker: The worker object. Used for multithreading.
    :param train_data: The data used for training.
    :param id_num: The id number to assign the neural network.
    :param column_train_test_split: The split between how many columns are used for training and how many are left for
                                    potential labels.
    :param batch_size: The number of rows to process at a time. Note this is system memory dependant.
    :param num_epochs: The number of epochs to be used in the training.
    :param num_gpus: The number of GPUs to distribute the training over.
    :param layers_dict: Dictionary of hidden layers to use in the ensemble networks
                        mapped as: "layer name" : number of units in the layer
    :return: The neural network object.
    """


    cols = train_data.shape[1]  # The number of columns in the data set

    unused_labels = list(range(0, cols - 1))  # A list of column indices that hasn't been used

    # If all of the columns have been used, reuse them all.
    if not unused_labels:
        unused_labels = list(range(0, cols))

    # Randomly pick one of those to be the new label
    label_index = unused_labels[random.randint(0, len(unused_labels) - 1)]
    # Remove the label from the index
    unused_labels.remove(label_index)
    # Retrieve all of the data from the column index
    net_training_label = train_data[:, label_index]

    # Generate a list of booleans indicating columns to keep (True) and to ignore (False)
    mask = np.random.rand(cols-1) < column_train_test_split

    # The columns used for training the neural nets
    remaining_cols = train_data
    remaining_cols = np.delete(remaining_cols, label_index, 1)

    net_training_columns = remaining_cols[:, mask]

    # Build and fit the net.
    if layers_dict is not None:
        layers_dict['inputLayer'] = net_training_columns.shape[1]
    else:
        layers_dict = {'inputLayer': net_training_columns.shape[1], 'hidden1': 10, 'hidden2': 4, 'output': 1}

    # Boolean array that includes a FALSE at the original index of the temp label
    total_mask = np.insert(mask, label_index, False)

    net = en.FeedForwardEnsembleNet(net_training_columns.shape[1],
                                    layers_dict, total_mask,
                                    label_index, name="ENet_" + str(id_num),
                                    batch_size=batch_size)
    net.build()

    if num_gpus:
        # The beginning of the multi-gpu implementation
        parallel_net = ks.utils.multi_gpu_model(net, num_gpus)
        parallel_net.fit(net_training_columns, net_training_label, 10)

        # Add the net to the ensemble
        return parallel_net

    else:
        net.fit(net_training_columns, net_training_label, num_epoch=num_epochs)
        return net
