# Adapter.py
# Author:       Luke Buquicchio

# Date Created: 1/19/18
# Description:
#   This class creates the final layers of the ensemble to aggregate their penultimate layers. Instead of coding
# this as one continuous neural net that must wait for every member of the ensemble to finish as well as fix each nodes'
# attributes, subsetPredictionNet will produce a matrix where each row is the concatenation of the penultimate layers'
# outputs for a given data point. This will then be passed to the adapter for processing.


from keras.models import Sequential
from keras.layers import Dense
from keras.utils import multi_gpu_model
from keras import metrics


class Adapter:

    def __init__(self, columns, verbose=0, loss='categorical_crossentropy', optimizer='adam',
                 metrics=['accuracy', metrics.mae, metrics.cosine], activation='relu', batch_size=64, hidden_units={'hidden_1': 30},
                 num_gpus=0, out_shape=1):
        """
        The __init__ function will serve to build and compile the model.
        :param columns: The number of columns to accurately build the input layer of the model.
        :param verbose: Whether or not to show the progress of the method.
        :param batch_size: The number of rows to process at a time. Note this is system memory dependant.
        :param loss: The loss function to be used.
        :param optimizer: The optimizer to be used.
        :param metrics: The metrics to measure the model performance.
        :param activation: The activation function to be used.
        :param initializer: The kernel initializer to be used.
        :param hidden_units: The number of nodes in the hidden layer.
        :param num_gpus: The number of GPUs for the training and testing to be distributed across.
        """
        self.columns = columns
        self.verbose = verbose
        self.batch_size = batch_size
        self.activation = activation
        self.model = Sequential()
        self.build(columns, hidden_units, out_shape=out_shape)
        self.model.compile(loss=loss, optimizer=optimizer, metrics=metrics)
        if num_gpus:
            # The beginning of the multi-gpu implementation
            self.model = multi_gpu_model(self, num_gpus)
            self.multi_gpu = True

        else:
            self.multi_gpu = False

    def build(self, in_shape, hidden_dict, out_shape):
        """
        The function to build the model.
        :param in_shape: The input shape.
        :param hidden_dict: The dictionary object of the hidden layers.
        :param out_shape: The shape of the output.
        :return:
        """
        self.model.add(Dense(in_shape,
                             input_dim=in_shape,
                             bias_initializer='zeros',
                             kernel_initializer='random_uniform',
                             activation='relu',
                             name="adapter_input_layer"))
        for key in hidden_dict:
            self.model.add(Dense(hidden_dict[key],
                                 activation='relu',
                                 bias_initializer='zeros',
                                 kernel_initializer='random_uniform',
                                 name="adapter_" + key))
        self.model.add(Dense(out_shape,
                             activation=self.activation,
                             bias_initializer='zeros',
                             kernel_initializer='random_uniform',
                             name="adapter_output_layer"))

    def fit(self, x_train, y_train, epochs=128):
        """
        The wrapper method for Keras' fit method. It will handle both a traditional modela as well as a multi-GPU model.
        :param x_train: The data to be used for training the model.
        :param y_train: The data labels to be used for the supervised part of the learning process.
        :param epochs: The number of iterations to use for training.
        """
        if self.multi_gpu:
            history = self.model.fit(x_train, y_train, epochs, validation_split=.2)
        else:
            history = self.model.fit(x_train, y_train, epochs, self.batch_size, verbose=self.verbose, validation_split=.2)

        return history

    def evaluate(self, x_test, y_test):
        """
        The wrapper method for Keras' evaluate method.
        :param x_test: The data to be used for testing the model.
        :param y_test: The data labels to be used to compare the y_hat to.
        :return: The metrics specified in __init__
        """
        return self.model.evaluate(x_test, y_test, self.batch_size, verbose=self.verbose)

    def predict(self, x_test):
        """
        The wrapper for Keras' predict method.
        :param x_test: The data to be used for testing the model.
        :return: The predictions for each row in x_test.
        """
        return self.model.predict(x_test, self.batch_size)

    def save(self, filepath):
        """
        The wrapper for Keras' save method.
        :param filepath: The file path for the model to be saved at
        :return:
        """
        self.model.save(filepath)
