# subsetPredictionNet
# Author:       Everett Harding, Luke Buquicchio, Juan Luis Herrero Estrada

# Date Created: 1/19/18
# Description:
#   This is the class to create the subset nets used in the bottom half of our EVS framework.
# The class handles selecting a subset of columns to use as features and a separate column to use as
# its target variable.

import keras as ks


class ConvolutionalEnsembleNet:
    """A convolutional network for use in ensemble layer of the framework with time series data."""

    def __init__(self, filters=1, kernel_size=10, strides=1, padding='valid', dilation_rate=1, activation=None,
                 use_bias=True, kernel_initializer='glorot_uniform', bias_initializer='zeros', name="conv"):
        """
        The __init__ function will serve to build a member of our ensemble.
        TODO Parametrize specifics of convolution.
        TODO Finish the implementation.
        :param filters: Integer, the dimensionality of the output space (i.e. the number output of filters in the
                        convolution).
        :param kernel_size: An integer or tuple/list of a single integer, specifying the length of the 1D convolution
                        window.
        :param strides: An integer or tuple/list of a single integer, specifying the stride length of the convolution.
                        Specifying any stride value != 1 is incompatible with specifying any dilation_rate value != 1.
        :param padding: One of "valid", "causal" or "same" (case-insensitive). "valid" means "no padding". "same"
                        Results in padding the input such that the output has the same length as the original input.
                        "causal" Results in causal (dilated) convolutions, e.g. output[t] does not depend on
                        input[t+1:]. Useful when modeling temporal data where the model should not violate the temporal
                        order.
        :param dilation_rate: An integer or tuple/list of a single integer, specifying the dilation rate to use for
                              dilated convolution. Currently, specifying any dilation_rate value != 1 is incompatible
                              with specifying any strides value != 1.
        :param activation: Activation function to use. If you don't specify anything, no activation is
                           applied (ie. "linear" activation: a(x) = x).
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param kernel_initializer: Initializer for the kernel weights matrix.
        :param bias_initializer: Initializer for the bias vector.
        """
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.dilation_rate = dilation_rate
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.name = name

        self.model = ks.models.Sequential()
        self.layers = {
            'input':
            ks.layers.Conv1D(
                filters=filters,
                kernel_size=kernel_size,
                strides=strides,
                padding=padding,
                dilation_rate=dilation_rate,
                activation=activation,
                use_bias=use_bias,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer)
        }
        self.model.add(self.layers['input'])
        # maintain a string representation of the model's state
        # before model is built: "model needs to be built"
        # before model is trained: "model has yet to be trained on any data"
        # after training:
        self.modelSummary = self.name + " needs to be built"


    def build(self):
        """
        Build the model, based on the layers specified. Return the finished model.
        :return:
        """
        # TODO Finish the method
        return None

    def train(self):
        """
        Train the model. Return the trained model.
        :return:
        """
        # TODO Finish the method
        return None

    def test(self):
        """
        Test the model.
        # TODO Finish the method
        :return:
        """
        return None

    def predict(self, data):
        """
        Apply the trained netowrk to a given example.
        # TODO Finish the method
        :return:
        """
        return None


class FeedForwardEnsembleNet:
    """A simple FFNN for use in ensemble layer of the framework."""

    """ a simple FFNN for use in ensemble layer of the framework"""
    def __init__(self, input_dim, layers_units, feature_mask, target_feature, loss="mse",
                 name="ffn", batch_size=100, output_shape=1, learning_rate=0.01, activation="relu", initializer="uniform"):
        """
        The __init__ function will serve to build a member of our ensemble.
        :param input_dim: The number of dimensions of the input layer.
        :param layers_units: A dictionary of the form {"layer name": # of units in the layer}
                             where layer name is a string, and # of units an integer
                             the first layer must always be mapped as "inputLayer".
        :param feature_mask: The boolean list of columns used.
        :param loss: The loss function to use for the ensemble member.
        :param name: The name of the model (for identification during testing & tuning), a string.
        :param batch_size: The size of the batches, an integer.
        :param output_shape: The size of the output layer (1 for continuous functions, >1 for classification).
        :param learning_rate: The rate of change for gradient descent.
        :param activation: The activation function.
        :param initializer: The initializer.
        """
        self.name = name
        self.layers_units = layers_units
        self.batch_size = batch_size
        self.output_shape = output_shape
        self.learning_rate = learning_rate
        self.feature_mask = feature_mask
        self.activation = activation
        self.initializer = initializer
        self.target_feature = target_feature
        self.modelAccuracy = 0

        self.model = ks.models.Sequential()
        self.layers = {
            'input':
            ks.layers.Dense(
                layers_units['inputLayer'],
                input_dim=input_dim,
                kernel_initializer="uniform",
                activation="relu")
        }
        self.model.add(self.layers['input'])
        del self.layers_units['inputLayer']
        # maintain a string representation of the model's state
        # before model is built: "model needs to be built"
        # before model is trained: "model has yet to be trained on any data"
        # after training:
        self.build()
        self.model.compile(optimizer='adam', loss=loss, metrics=['accuracy'])

    def build(self):
        """
        Assemble the layers of the model.
        Initialized with uniform distribution and activation function relu.

        :return: None
        """
        for key in self.layers_units:
            units = self.layers_units[key]
            self.layers[key] = ks.layers.Dense(units,
                                               kernel_initializer=self.initializer,
                                               activation=self.activation,
                                               name=self.name+"_"+key)
            self.model.add(self.layers[key])

        self.modelSummary = self.name + " is built, but has yet to be trained on any data"

    def fit(self, feature_subset, target_feature, num_epoch=25):
        """
        Fit the model based on the given feature set and targets.

        :param feature_subset: Training set containing a subset of all features available, an ndarray.
        :param target_feature: Training set of labels, where the label is a feature of the dataset, a 1d vector.
        :param num_epoch: Number of training epochs, defaulted to 1, an integer.
        :return: None
        """
        sgd = ks.optimizers.SGD(lr=0.01)
        # print("Compiling & Fitting Model", self.name + "...")
        self.model.fit(feature_subset,
                       target_feature,
                       epochs=num_epoch,
                       batch_size=self.batch_size,
                       verbose=1)

    def evaluate(self, feature_subset, target_feature):
        """
        Evaluate the model on the evaluate set.

        :param feature_subset: Test set, formed of the same feature columns as the training set, an ndarray.
        :param target_feature: Labels for the evaluate set, the same feature as for the training set, a vectory.
        :return: None
        """
        loss, accuracy = self.model.evaluate(x=feature_subset, y=target_feature, verbose=1)

        # store results in modelSummary, print them out
        self.modelSummary = "\n" + self.name + " shape: " + str([key + ": " + str(self.layers_units[key]) for key in self.layers_units])\
                            + "\nMetrics:\n\tLoss: " + str(loss) + "\n\tAccuracy: " + str(accuracy * 100) + "\n\n"
        return loss, accuracy

    def predict(self, feature_subset):
        """
        Generate the output of the model.

        :param feature_subset: Set to generate predictions from, formed of the same columns as training set, an ndarray.
        :return: Model's transformation of the feature_subset, a vector.
        """

        return self.model.predict(feature_subset, verbose=0)


    def save(self, filepath):
        """
        The wrapper for Keras' save method.
        :param filepath: The file path for the model to be saved at
        :return:
        """
        self.model.save(filepath)
